/*
 [The "BSD licence"]
 Copyright (c) 2005 Jean Bovet, http://www.cs.usfca.edu/~jbovet/
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
 derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
														   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
														   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "Sniffer.h"
#import "EthernetPacket.h"

#include <net/bpf.h>
#include <fcntl.h>
#include <sys/ioctl.h>

@implementation Sniffer

- (id)init
{
	if(self = [super init]) {
		mDelegate = NULL;
		mThreadRunning = NO;
		mThreadRuns = NO;
		mQuiet = NO;
		mBPFSocket = -1;
	}
	return self;
}

- (void)setDelegate:(id)delegate
{
	mDelegate = delegate;
}

- (void)setBPFSocket:(int)socket
{
	mBPFSocket = socket;
}

#pragma mark -

- (void)start
{
	if(!mThreadRunning) {
		mThreadRunning = YES;
		[NSApplication detachDrawingThread:@selector(run) toTarget:self withObject:NULL];		
	}
}

- (void)stop
{
	if(mThreadRunning) {
		mThreadRunning = NO;
		while(mThreadRuns) {
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.01]];
		}		
	}
}

- (void)quiet:(BOOL)flag
{
	mQuiet = flag;
}

#pragma mark -

- (void)run
{	
	mThreadRuns = YES;
	
	// Set the socket to be non-blocking
	int err = fcntl(mBPFSocket, F_SETFL, O_NONBLOCK);
	if (err != noErr)
    {
        fprintf(stderr, "fcntl returned error %d\n", err);
        return;
    }
		
	// Get the size of the BPF buffer
	u_int		blen;
    err = bpf_get_blen(mBPFSocket, &blen);
    if (err < 0)
    {
        fprintf(stderr, "bpf_get_blen returned error %d\n", err);
        return;
    }
	
	// Allocate the BPF buffer
	char *buffer = (char*)malloc(blen);
	if (err < 0)
    {
        fprintf(stderr, "malloc returned error %d\n", err);
        return;
    }
		
	int i;
	while(mThreadRunning) {
		// Read any data available
        int n = read(mBPFSocket, buffer, blen);
        if (n < 0) {
            fprintf(stderr, "read failed with error %d\n", errno);
			break;
        } else if (n > 0 && !mQuiet) {
			char *p = buffer;
			// Multiple packets can be found in the buffer for each read
			while(p < buffer+n) {
				struct bpf_hdr *header = (struct bpf_hdr*)p;
				
				/*fprintf(stderr, "n %d, blen %d, caplen %4d, datalen %4d, hdrlen %4d\n", n, blen,
						header->bh_caplen,
						header->bh_datalen,
						header->bh_hdrlen);*/
				
				// Send the new packet to the delegate
				EthernetPacket *packet = [EthernetPacket packetWithData:[NSData dataWithBytes:p+header->bh_hdrlen length:header->bh_caplen]];
				NSDate *date = [NSDate dateWithTimeIntervalSince1970:header->bh_tstamp.tv_sec];
				[packet setTime:[NSString stringWithFormat:@"%@.%d", [date descriptionWithCalendarFormat:@"%H:%M:%S" timeZone:nil locale:nil], header->bh_tstamp.tv_usec]];
				[mDelegate performSelector:@selector(snifferPacketArrived:) withObject:packet];
				
				// Jump to the next packet (see man page of bpf)
				p += BPF_WORDALIGN(header->bh_hdrlen+header->bh_caplen);				
			}			
        } else {
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.01]];
		}
	}
		
	free(buffer);
	
	mThreadRuns = NO;
}

- (unsigned)packetsReceived
{	
	if(mBPFSocket > -1) {
		struct bpf_stat stat;	
		ioctl(mBPFSocket, BIOCGSTATS, &stat);
		return stat.bs_recv;		
	} else 
		return 0;
}

- (unsigned)packetsDropped
{
	if(mBPFSocket > -1) {
		struct bpf_stat stat;	
		ioctl(mBPFSocket, BIOCGSTATS, &stat);
		return stat.bs_drop;			
	} else
		return 0;
}

@end
